import numpy as np

from checkfield import checkfield
from fielddisplay import fielddisplay
from MatlabFuncs import *
from WriteData import WriteData

class stochasticforcing(object):
    """STOCHASTICFORCING class definition

    Usage:
        stochasticforcing = stochasticforcing()
    """

    def __init__(self, *args):  # {{{
        self.isstochasticforcing = 0
        self.fields              = np.nan
        self.defaultdimension    = 0
        self.default_id          = np.nan
        self.covariance          = np.nan
        self.randomflag          = 1

        if len(args) == 0:
            self.setdefaultparameters()
        else:
            error('constructor not supported')

    def __repr__(self):  # {{{
        s = '   stochasticforcing parameters:\n'
        s += '{}\n'.format(fielddisplay(self, 'isstochasticforcing', 'is stochasticity activated?'))
        s += '{}\n'.format(fielddisplay(self, 'fields', 'fields with stochasticity applied, ex: [\'SMBautoregression\'], or [\'FrontalForcingsRignotAutoregression\']'))
        s += '{}\n'.format(fielddisplay(self, 'defaultdimension', 'dimensionality of the noise terms (does not apply to fields with their specific dimension)'))
        s += '{}\n'.format(fielddisplay(self, 'default_id', 'id of each element for partitioning of the noise terms (does not apply to fields with their specific partition)'))
        s += '{}\n'.format(fielddisplay(self, 'covariance', 'covariance matrix for within- and between-fields covariance (units must be squared field units)'))
        s += '{}\n'.format(fielddisplay(self, 'randomflag', 'whether to apply real randomness (true) or pseudo-randomness with fixed seed (false)'))
        s += 'Available fields:\n'
        s += '   DefaultCalving\n'
        s += '   SMBautoregression\n'
        s += '   FrontalForcingsRignotAutoregression (thermal forcing)\n'
        return s
    #}}}

    def setdefaultparameters(self):  # {{{
        # Type of stabilization used
        self.isstochasticforcing = 0 # stochasticforcing is turned off by default
        self.fields              = [] # Need to initialize to list to avoid "RuntimeError: object of type 'float' has no len()" on import of class
        self.randomflag          = 1 # true randomness is implemented by default
        return self
    #}}}

    def checkconsistency(self, md, solution, analyses):  # {{{
        # Early return
        if not self.isstochasticforcing:
            return md

        num_fields  = numel(self.fields)

        #Check that covariance matrix is positive definite
        try:
            np.linalg.cholesky(self.covariance);
        except:
            error('md.stochasticforcing.covariance is not positive definite');

        # Check that all fields agree with the corresponding md class
        checkdefaults = False
        for field in self.fields:
            if (contains(field, 'SMB')):
                if not (type(md.smb) == field):
                    error('md.smb does not agree with stochasticforcing field {}'.format(field))
            if (contains(field, 'frontalforcings')):
                if not (type(md.frontalforcings) == field):
                    error('md.frontalforcings does not agree with stochasticforcing field {}'.format(field))
            #Checking for specific dimensions
            if (field!='SMBautoregression' or field!='FrontalForcingsRignotAutoregression'):
                checkdefaults = True #field with non-specific dimensionality

        #Retrieve sum of all the field dimensionalities
        size_tot   = self.defaultdimension*num_fields;
        indSMBar = -1; #about to check for index of SMBautoregression
        indTFar  = -1; #about to check for index of FrontalForcingsRignotAutoregression
        if ('SMBautoregression' in self.fields):
            size_tot = size_tot-self.defaultdimension+md.smb.num_basins
            indSMBar = np.where(self.fields=='SMBautoregression')[0][0]
        if ('FrontalForcingsRignotAutoregression' in self.fields):
            size_tot = size_tot-self.defaultdimension+md.frontalforcings.num_basins
            indSMBar = np.where(self.fields=='FrontalForcingsRignotAutoregression')[0][0]
        if (indSMBar!=-1 and indTFar!=-1):
            if((md.smb.ar_timestep!=md.frontalforcings.ar_timestep) and np.any(self.covariance[np.sum(self.dimensions[0:indSMBar]).astype(int):np.sum(self.dimensions[0:indSMBar+1]).astype(int),np.sum(self.dimensions[0:indTFar]).astype(int):np.sum(self.dimensions[0:indTFar+1]).astype(int)]!=0)):
                error('SMBautoregression and FrontalForcingsRignotAutoregression have different ar_timestep and non-zero covariance')

        md = checkfield(md, 'fieldname', 'stochasticforcing.isstochasticforcing', 'values', [0, 1])
        md = checkfield(md, 'fieldname', 'stochasticforcing.fields', 'numel', num_fields, 'cell', 1, 'values', supportedstochforcings())
        #md = checkfield(md, 'fieldname', 'stochasticforcing.dimensions', 'NaN', 1, 'Inf', 1, '>', 0, 'size', [num_fields]) # specific dimension for each field; NOTE: As opposed to MATLAB implementation, pass list
        md = checkfield(md, 'fieldname', 'stochasticforcing.covariance', 'NaN', 1, 'Inf', 1, 'size', [size_tot, size_tot]) # global covariance matrix
        md = checkfield(md, 'fieldname', 'stochasticforcing.randomflag', 'numel', [1], 'values', [0, 1])
        if (checkdefaults):
            md = checkfield(md, 'fieldname', 'stochasticforcing.defaultdimension', 'numel', 1, 'NaN', 1, 'Inf', 1, '>', 0)
            md = checkfield(md, 'fieldname', 'stochasticforcing.default_id', 'Inf', 1, '>=', 0, '<=', self.defaultdimension, 'size', [md.mesh.numberofelements,1])
        return md
    # }}}

    def extrude(self, md):  # {{{
        # Nothing for now
        return self
    # }}}

    def marshall(self, prefix, md, fid):  # {{{
        yts = md.constants.yts
        num_fields = len(self.fields)

        WriteData(fid, prefix, 'object', self, 'fieldname', 'isstochasticforcing', 'format', 'Boolean')
        if not self.isstochasticforcing:
            return md
        else:
            #Retrieve dimensionality of each field
            dimensions = self.defaultdimension*np.ones(num_fields)
            for ind,field in enumerate(self.fields):
                #Checking for specific dimensions
                if (field=='SMBautoregression'):
                    dimensions[ind] = md.smb.num_basins
                if (field=='FrontalForcingsRignotAutoregression'):
                    dimensions[ind] = md.frontalforcings.num_basins

            # Scaling covariance matrix (scale column-by-column and row-by-row)
            scaledfields = ['DefaultCalving','SMBautoregression'] # list of fields that need scaling * 1/yts
            tempcovariance = np.copy(self.covariance)
            for i in range(num_fields):
                if self.fields[i] in scaledfields:
                    inds = range(int(np.sum(self.dimensions[0:i])), int(np.sum(self.dimensions[0:i + 1])))
                    for row in inds: # scale rows corresponding to scaled field
                        tempcovariance[row, :] = 1 / yts * tempcovariance[row, :]
                    for col in inds: # scale columns corresponding to scaled field
                        tempcovariance[:, col] = 1 / yts * tempcovariance[:, col]
            #Set dummy default_id vector if defaults not used
            if np.isnan(self.default_id):
               self.default_id = np.zeros(md.mesh.numberofelements)
            WriteData(fid, prefix, 'data', num_fields, 'name', 'md.stochasticforcing.num_fields', 'format', 'Integer')
            WriteData(fid, prefix, 'object', self, 'fieldname', 'fields', 'format', 'StringArray')
            WriteData(fid, prefix, 'data', 'dimensions', 'name', 'md.stochasticforcing.dimensions', 'format', 'IntMat')
            WriteData(fid, prefix, 'object', self, 'fieldname', 'default_id', 'format', 'IntMat') #12Nov2021 make sure this is zero-indexed!
            WriteData(fid, prefix, 'object', self, 'fieldname', 'defaultdimension', 'format', 'Integer') 
            WriteData(fid, prefix, 'data', tempcovariance, 'name', 'md.stochasticforcing.covariance', 'format', 'DoubleMat')
            WriteData(fid, prefix, 'object', self, 'fieldname', 'randomflag', 'format', 'Boolean')
    # }}}

def supportedstochforcings():
    """ Defines list of fields supported  by the class stochasticforcings
    """
    return [
        'DefaultCalving',
        'FrontalForcingsRignotAutoregression',
        'SMBautoregression'
    ]
